/* logoutd.c
 * Written by Edsel Adap. <edsel.adap@canada.sun.com>
 * Sun Microsystems Developer Support Center
 * June 25, 1996
 *
 * This program is intended to be run as a background process.
 * It takes one command line argument -- an integer.
 * This integer specifies the number of minutes this program sleeps
 * before waking up to check for idle users.
 * An idle user is defined as a user with no processes other than the
 * shell running and has not had any terminal I/O in the interval
 * specified in the command line.
 *
 * This program was written under Solaris 2.5.  It may or may not work
 * on other versions of Solaris or other Unix or Unixlike systems.
 *
 * Usage: logoutd n
 *        where n is the interval in minutes between checks
 *
 * Algorithm:
 * 1. Scan utmp for users logged in with no activity for n minutes
 * 2. Scan /proc file system for processes associated with the idle terminal
 * 3. If no processes other than the shell is found, kill the shell by
 *    sending a SIGHUP signal to it.
 * 4. Sleep for n minutes
 * 5. wake up and start at step 1.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <utmp.h>
#include <time.h>
#include <sys/types.h>
#include <sys/signal.h>
#include <sys/fault.h>
#include <sys/syscall.h>
#include <sys/procfs.h>
#include <sys/stat.h>
#include <dirent.h>
#include <fcntl.h>
#include <pwd.h>

/* Globals */
time_t now;
char *me;
time_t idletime,idle;

struct proctab {
	char name[256];
	pid_t pid;
} proclist[65536];

/* Function Prototypes */
void logoutd(void);
static time_t findidle(char *);
int get_proc_list(char *,dev_t, int *);

/* findidle()
 * Function which takes a character string as an argument
 * This string contains the name of a user's tty.
 * Returns the number of minutes the tty has been idle
 */
static time_t findidle(char *devname)
{
    struct stat stbuf;
    time_t lastaction, diff;
    char ttyname[64];

	/* Build the pathname to the tty device */
    strcpy(ttyname, "/dev/");
    strcat(ttyname, devname);

	/* Find out the last time the device file was accessed */
    if (stat(ttyname, &stbuf) != -1) {
        lastaction = stbuf.st_atime;
        diff = ((now - lastaction) + 30 )/60; /* convert to minutes */
        if (diff < 0)
            diff = 0;
    } else
        diff = 0;

    return (diff);
}

/*  main()
 *  All we do here is check for arguments call logoutd() and then sleep
 *  loop with sleep() and logoutd() calls inside the infinite loop
 */ 
main (int argc, char **argv){

	me=argv[0]; /* our program name */

	if ( argc != 2 ) { 	/* require an argument */
		fprintf(stderr,"Usage: %s n\n",me);
		exit(1);
	}

	idletime=atoi(argv[1]); /* determine how long we are to sleep */

	if(idletime==0){ 		/* time specified must be greater than 0 */
		fprintf(stderr,"Usage: %s n\nWhere n is the number of minutes\na process must be idle before\nit can be flagged as idle.\n",me);
		exit(1);
	}

	if(getuid()) { /* are we root? */
		fprintf (stderr,"You must be root to run this program\n");
		exit(1);
	}

	/* We really don't need standard in and standard out */
	close(0);
	close(1);

	logoutd(); /* initial call to logoutd */

	while (1) { /* loop forever */
		sleep(idletime*60); /* sleep for idletime number of minutes */
		logoutd();			/* check for idle users when we wake up */
	}
}

/* logoutd()
 * Checks for idle users using findidle
 * Collects process information about the tty
 * kills shell if it is the only process using that tty.
 */
void logoutd(void){

	struct utmp *utmpbegin, *ut;
	struct stat	sbuf;
	unsigned utmpend, size;
	int fd,nprocs,i;
	char devname[14];
	
	/* get info about the /var/adm/utmp file */
	if (stat(UTMP_FILE, &sbuf)) {
		fprintf(stderr,"%s: Unable to stat utmp file %s\n",me,UTMP_FILE);
		exit(1);
	}

	size = (unsigned) sbuf.st_size;
	if ( (ut = malloc(size)) == NULL ) { /* allocate memory for utmp file */
		fprintf(stderr,"%s: malloc error\n",me);
		exit(1);
	}

	if ( (fd = open(UTMP_FILE,O_RDONLY)) == -1) { /* open utmp for reading */
		fprintf(stderr,"%s: Unable to open utmp file %s for reading\n",
				me,UTMP_FILE);
		exit(1);
	}

	if (read (fd, (char *)ut, size) == -1) { /* read the whole utmp file */
		fprintf(stderr,"%s: Error reading utmp file\n",me,UTMP_FILE);
		exit(1);
	}
	utmpbegin = ut; 	/* ptr to start of utmp data */
	utmpend = (unsigned) ut + size; /* ptr to end of utmp data */
	close(fd);

	time(&now);	 /* get current time */

	/* loop through every entry in the utmp file */
	for (ut = utmpbegin; ut < (struct utmp *)utmpend; ut++){

		/* we are only interested in user processes */
		if (ut->ut_type == USER_PROCESS) {
			sprintf(devname,"/dev/%s",ut->ut_line); /* build device pathname */
			stat(devname, &sbuf); /* get info on device file */
			idle = findidle(ut->ut_line); 	/* find user idle time in minutes */
			if (idle >= idletime){
				nprocs=0; /* always set the number of procs initially to 0 */
				/* get a list of processes using the idle tty */
				if(get_proc_list(ut->ut_user, sbuf.st_rdev, &nprocs)){
					fprintf(stderr,"get_proc_list error\n");
				}else{
					if(nprocs == 1){
						/* we only have one process on the idle tty */
						/* presumably this is the user's shell */
						kill(proclist[0].pid,SIGHUP); /* send a sighup */
					}
					/* if we have more than 1 processes then chances are
					   the process is taking a long time to send any
					   output to the tty.  The user may be running a cpu
					   intensive job.  We don't want to kill that.
                     */
				}
			}
		}
	}
	free(ut);
}

/* get_proc_list()
 * function parameters:
 *		char *owner   -> string.  the username whose processes we are
 *                       interested in
 *      dev_t tty     -> the tty we are interested in (the idle one)(
 *      int *num      -> counter for the number of processes we found
 *
 * scans /proc to get a list of all the processes on the system
 * selects the ones whose user and tty matches the one passed to the function
 * stores found process in the global variable proclist
 * returns -1 if there is a problem ... should never happen but we never know
 * returns 0 if successful
 */
int get_proc_list(char *owner,dev_t tty, int *num){

	DIR *procd;

	struct dirent *procf;
	struct prpsinfo psinfo;
	int p,i,owner_uid,pid;
	char *user;
	char filename[14];
	struct passwd *pwd;
	extern int errno;
	
	pwd=getpwnam(owner); /* we're really interested with the uid of the owner */
	if(pwd==NULL){
		fprintf(stderr,"Problem finding matching uid for %s\n",owner);
		fprintf(stderr,"Perhaps you have a corrupted %s file\n",UTMP_FILE);
		return(-1);
	}
	owner_uid=pwd->pw_uid;

	procd = opendir("/proc"); /* lets open /proc */
	i=0;
	while( (procf = readdir(procd)) != NULL ) {
		if(i++<2) continue; /* skip the first two entries -- "." and ".." */
		pid=atoi(procf->d_name); /* convert the filename (pid) to an integer */
		sprintf(filename,"/proc/%s",procf->d_name);
		p=open(filename,O_RDONLY);	/* lets open the process for reading */

		if (p == -1 ) {
			fprintf(stderr,"Unable to open process %d\n",atoi(procf->d_name));
			fprintf(stderr,"error %d\n",errno);
			continue; /* if we have a problem opening this process lets
                         skip to the next process */
		}
		if ( ioctl(p,PIOCPSINFO,&psinfo) == -1 ) { /* get PS info */
			fprintf(stderr, "%s: problem reading process number %d\n",me, pid);
			fprintf(stderr, "    Error number %d\n",errno);
		}else{
			/* if owner and tty matches we store the process name and pid */
			/* really we only need the pid, but for testing purposes */
			/* the name might be useful */
			if(owner_uid == psinfo.pr_uid && tty == psinfo.pr_ttydev){
				strcpy(proclist[*num].name,psinfo.pr_fname);
				proclist[*num].pid = psinfo.pr_pid; 
				(*num)++; /* increment the count of processes found */
			}
		}	
		close(p);
	}
	return(0);
}
